@@ -1,5 +1,6 @@
 import dataclasses
-import itertools
+import enum
+import typing
 from collections.abc import Iterable, Iterator
 from functools import singledispatchmethod
 from typing import Optional, TypeVar, Union
@@ -7,7 +8,6 @@ from typing import Optional, TypeVar, Union
 from doc_printer import (
     Doc,
     DocLike,
-    Empty,
     Fail,
     Line,
     Space,
@@ -28,86 +28,71 @@ from doc_printer.doc import splat
 from tree_sitter_talon import (
     Node,
     TalonAction,
-    TalonAnd,
     TalonArgumentList,
-    TalonAssignment,
+    TalonAssignmentStatement,
     TalonBinaryOperator,
     TalonBlock,
     TalonCapture,
     TalonChoice,
-    TalonCommand,
+    TalonCommandDeclaration,
     TalonComment,
-    TalonContext,
+    TalonDeclaration,
+    TalonDeclarations,
     TalonEndAnchor,
-    TalonExpression,
+    TalonExpressionStatement,
     TalonFloat,
     TalonIdentifier,
     TalonImplicitString,
-    TalonIncludeTag,
     TalonInteger,
     TalonInterpolation,
     TalonKeyAction,
+    TalonKeyBindingDeclaration,
     TalonList,
     TalonMatch,
-    TalonNot,
-    TalonNumber,
+    TalonMatches,
+    TalonMatchModifier,
     TalonOperator,
     TalonOptional,
-    TalonOr,
     TalonParenthesizedExpression,
     TalonParenthesizedRule,
-    TalonRegexEscapeSequence,
     TalonRepeat,
     TalonRepeat1,
     TalonRule,
     TalonSeq,
-    TalonSettings,
+    TalonSettingsDeclaration,
     TalonSleepAction,
     TalonSourceFile,
     TalonStartAnchor,
+    TalonStatement,
     TalonString,
     TalonStringContent,
     TalonStringEscapeSequence,
+    TalonTagImportDeclaration,
+    TalonUnaryOperator,
     TalonVariable,
     TalonWord,
 )
 
-TalonBlockLevelMatch = Union[
-    TalonAnd,
-    TalonNot,
-    TalonMatch,
-    TalonOr,
-]
+from .assert_equivalent import *
 
 TalonBlockLevel = Union[
     TalonSourceFile,
-    TalonContext,
-    TalonIncludeTag,
-    TalonSettings,
-    TalonCommand,
+    TalonMatches,
+    TalonMatch,
+    TalonDeclarations,
+    TalonDeclaration,
     TalonBlock,
-    TalonAssignment,
-    TalonExpression,
+    TalonStatement,
     TalonComment,
 ]
 
 NodeVar = TypeVar("NodeVar", bound=Node)
 
 
-def is_short_command(node: TalonCommand) -> bool:
-    return len(node.children) + len(node.script.children) == 1
-
-
-def block_with_comments(
-    comments: Iterable[TalonComment], block: TalonBlock
-) -> TalonBlock:
-    return TalonBlock(
-        text=block.text,
-        type_name=block.type_name,
-        start_position=block.start_position,
-        end_position=block.end_position,
-        children=[*comments, *block.children],
-    )
+class EmptyMatchContext(enum.IntEnum):
+    Show = 0
+    Keep = 1
+    Hide = 2
 
 
 @dataclasses.dataclass
@@ -115,12 +100,20 @@ class TalonFormatter:
     indent_size: int
     align_match_context: Union[bool, int]
     align_short_commands: Union[bool, int]
-    blank_line_after_match_context: bool
+    empty_match_context: EmptyMatchContext
     format_comments: bool
     preserve_blank_lines_in_header: bool
     preserve_blank_lines_in_body: bool
     preserve_blank_lines_in_command: bool
 
+    @property
+    def show_empty_match_context(self) -> bool:
+        return self.empty_match_context is EmptyMatchContext.Show
+
+    @property
+    def keep_empty_match_context(self) -> bool:
+        return self.empty_match_context is EmptyMatchContext.Keep
+
     @singledispatchmethod
     def format(self, node: Node) -> Doc:
         """
@@ -131,13 +124,11 @@ class TalonFormatter:
             node,
             (
                 TalonSourceFile,
-                TalonContext,
-                TalonIncludeTag,
-                TalonSettings,
-                TalonCommand,
+                TalonMatches,
+                TalonDeclaration,
                 TalonBlock,
-                TalonAssignment,
-                TalonExpression,
+                TalonStatement,
+                TalonComment,
             ),
         ):
             return cat(self.format_lines(node))
@@ -151,24 +142,9 @@ class TalonFormatter:
         """
         if isinstance(node, TalonComment):
             yield self.format(node)
-        elif isinstance(node, (TalonAnd, TalonNot, TalonMatch, TalonOr)):
-            yield from self.format_lines_match(node, under_and=False, under_not=False)
         else:
             raise TypeError(type(node))
 
-    @singledispatchmethod
-    def format_lines_match(
-        self,
-        match: TalonBlockLevelMatch,
-        *,
-        under_and: bool,
-        under_not: bool,
-    ) -> Iterator[Doc]:
-        """
-        Format any match statement or comment as a series of lines.
-        """
-        raise TypeError(type(match))
-
     def format_children(self, children: Iterable[Node]) -> Iterator[Doc]:
         for child in self.store_comments_with_type(children, node_type=Node):
             if isinstance(child, Iterable):
@@ -183,17 +159,17 @@ class TalonFormatter:
     @format_lines.register
     def _(self, node: TalonSourceFile) -> Iterator[Doc]:
 
-        # Used to emit the context header separator.
+        # Used to emit the match context separator.
         in_header: bool = True
 
         # Used to buffer comments to ensure that they're split correctly
         # between the header and body.
-        comment_buffer: list[Doc] = []
+        match_context_comment_buffer: list[Doc] = []
 
-        def clear_comment_buffer() -> Iterator[Doc]:
-            if comment_buffer:
-                yield from comment_buffer
-                comment_buffer.clear()
+        def clear_match_context_comment_buffer() -> Iterator[Doc]:
+            if match_context_comment_buffer:
+                yield from match_context_comment_buffer
+                match_context_comment_buffer.clear()
 
         if self.align_short_commands is True:
             # Used to buffer short commands to group them as tables.
@@ -211,70 +187,75 @@ class TalonFormatter:
         # Used to insert blank lines.
         previous_line: int = 0
 
-        for child in node.children:
+        # Iterate over children, flatten any TalonDeclarations node
+        def children() -> Iterator[
+            typing.Union[TalonDeclaration, TalonMatches, TalonComment]
+        ]:
+            for child in node.children:
+                if isinstance(child, TalonDeclarations):
+                    yield from child.children
+                else:
+                    yield child
+
+        for child in children():
             extra_blank_line: bool = child.start_position.line - previous_line >= 2
 
+            # buffer comments in match context
             if in_header and isinstance(child, TalonComment):
-                # NOTE: buffer comments in context header
                 if self.preserve_blank_lines_in_body and extra_blank_line:
-                    comment_buffer.append(Line)
-                comment_buffer.append(self.format(child))
-            elif isinstance(child, TalonContext):
-                # NOTE: format context header
-                assert in_header
-                yield from clear_comment_buffer()
+                    match_context_comment_buffer.append(Line)
+                match_context_comment_buffer.append(self.format(child))
+
+            # format the .talon file match context
+            elif isinstance(child, TalonMatches):
+                assert in_header  # must still be in the header
+                yield from clear_match_context_comment_buffer()
                 yield from self.format_lines(child)
-            else:
-                # NOTE: first body-only node, end the context header
-                if in_header and isinstance(
-                    child, (TalonIncludeTag, TalonSettings, TalonCommand)
+                if (
+                    bool(child.children)
+                    or (child.is_explicit() and self.keep_empty_match_context)
+                    or self.show_empty_match_context
                 ):
                     yield Text("-") / Line
-                    # NOTE: no blank lines after "-" unless
-                    #       blank_line_after_match_context is set
-                    if self.blank_line_after_match_context:
-                        yield Line
-                    if comment_buffer:
-                        yield from itertools.dropwhile(
-                            lambda doc: doc is Line, clear_comment_buffer()
-                        )
-                    else:
-                        extra_blank_line = False
-                    in_header = False
+                in_header = False
 
+            # format the .talon file body
+            else:
+                # for dynamic alignment:
+                #   buffer short commands and clear the short command buffer
+                #   when anything other kind of node is encountered
                 if self.align_short_commands is True:
-                    if isinstance(child, TalonCommand) and is_short_command(child):
-                        # NOTE: buffer short command
+                    if isinstance(child, TalonCommandDeclaration) and child.is_short():
+                        if self.preserve_blank_lines_in_body and extra_blank_line:
+                            if not short_command_buffer:
+                                yield Line
                         short_command_buffer.extend(self.format_lines(child))
                     else:
-                        # NOTE: long command or other node, clear short command buffer
                         yield from clear_short_command_buffer()
                         if self.preserve_blank_lines_in_body and extra_blank_line:
                             yield Line
                         yield from self.format_lines(child)
+
+                # otherwise:
+                #   emit formatted nodes as they are encountered
                 else:
                     if self.preserve_blank_lines_in_body and extra_blank_line:
                         yield Line
                     yield from self.format_lines(child)
 
-            # Update previous line.
+            # update previous line
             previous_line = child.end_position.line
 
-        # NOTE: no body-only node, end the context header
-        if in_header:
-            yield Text("-") / Line
-            yield from clear_comment_buffer()
-
-        # NOTE: clear remaining short commands in buffer
+        # file ends with a short command, clear the short command buffer
         if self.align_short_commands is True:
             yield from clear_short_command_buffer()
 
     ###########################################################################
-    # Format: Context Header
+    # Format: Match Context
     ###########################################################################
 
     @format_lines.register
-    def _(self, node: TalonContext) -> Iterator[Doc]:
+    def _(self, node: TalonMatches) -> Iterator[Doc]:
 
         # Used to insert blank lines.
         previous_line: Optional[int] = None
@@ -292,39 +273,12 @@ class TalonFormatter:
             # Update previous line.
             previous_line = child.end_position.line
 
-    @format_lines_match.register
-    def _(self, match: TalonAnd, under_and: bool, under_not: bool) -> Iterator[Doc]:
-        for child in match.children:
-            if isinstance(child, TalonComment):
-                yield from self.format_lines(child)
-            else:
-                yield from self.format_lines_match(child, under_and, under_not)
-                under_and = True
-
-    @format_lines_match.register
-    def _(self, match: TalonNot, under_and: bool, under_not: bool) -> Iterator[Doc]:
-        for child in match.children:
-            if isinstance(child, TalonComment):
-                yield from self.format_lines(child)
-            else:
-                yield from self.format_lines_match(child, under_and, under_not=True)
-
-    @format_lines_match.register
-    def _(self, match: TalonOr, under_and: bool, under_not: bool) -> Iterator[Doc]:
-        for child in match.children:
-            if isinstance(child, TalonComment):
-                yield from self.format_lines(child)
-            else:
-                yield from self.with_comments(
-                    self.format_lines_match(child, under_and, under_not)
-                )
-
-    @format_lines_match.register
-    def _(self, match: TalonMatch, under_and: bool, under_not: bool) -> Iterator[Doc]:
-        self.assert_only_comments(match.children)
-        keywords = self.format_match_keywords(under_and, under_not)
-        key = self.format(match.key)
-        pattern = self.format(match.pattern)
+    @format_lines.register
+    def _(self, node: TalonMatch) -> Iterator[Doc]:
+        self.assert_only_comments(node.children)
+        keywords = self.format_match_modifiers(node.modifiers)
+        key = self.format(node.left)
+        pattern = self.format(node.right)
         if isinstance(self.align_match_context, bool):
             yield row(
                 keywords / key / ":",
@@ -339,49 +293,62 @@ class TalonFormatter:
                 min_col_widths=(self.align_match_context,),
             )
 
-    def format_match_keywords(self, under_and: bool, under_not: bool) -> Iterator[Doc]:
-        if under_and:
+    def format_match_modifiers(
+        self,
+        modifiers: typing.Sequence[TalonMatchModifier],
+    ) -> Iterator[Doc]:
+        if any(modifier.text == "and" for modifier in modifiers):
             yield Text("and")
             yield Space
-        if under_not:
+        if any(modifier.text == "not" for modifier in modifiers):
             yield Text("not")
             yield Space
 
     ###########################################################################
-    # Format: Tag Includes
+    # Format: Tag Import Declaration
     ###########################################################################
 
     @format_lines.register
-    def _(self, node: TalonIncludeTag) -> Iterator[Doc]:
+    def _(self, node: TalonTagImportDeclaration) -> Iterator[Doc]:
         self.assert_only_comments(node.children)
-        yield from self.with_comments("tag():" // self.format(node.tag) / Line)
+        yield from self.with_comments("tag():" // self.format(node.right) / Line)
 
     ###########################################################################
-    # Format: Settings
+    # Format: Settings Declaration
     ###########################################################################
 
     @format_lines.register
-    def _(self, node: TalonSettings) -> Iterator[Doc]:
-        block = self.get_node_with_type(node.children, node_type=TalonBlock)
-        block = block_with_comments(self.get_comments(), block)
+    def _(self, node: TalonSettingsDeclaration) -> Iterator[Doc]:
+        assert node.children is None
         yield "settings():" / nest(
             self.indent_size,
             Line,
-            self.format(block),
+            self.format(node.right),
         )
 
+    ###########################################################################
+    # Format: Key Bindings
+    ###########################################################################
+
+    @format_lines.register
+    def _(self, node: TalonKeyBindingDeclaration) -> Iterator[Doc]:
+        assert node.children is None
+        rule = self.format(node.left)
+        script = self.format(node.right)
+        yield from self.format_command(rule, script, node.is_short())
+
     ###########################################################################
     # Format: Commands
     ###########################################################################
 
     @format_lines.register
-    def _(self, node: TalonCommand) -> Iterator[Doc]:
-        rule = self.format(node.rule)
-
-        # Merge comments on this node into the block node.
-        block = block_with_comments(node.children, node.script)
-        script = self.format(block)
+    def _(self, node: TalonCommandDeclaration) -> Iterator[Doc]:
+        assert node.children is None
+        rule = self.format(node.left)
+        script = self.format(node.right)
+        yield from self.format_command(rule, script, node.is_short())
 
+    def format_command(self, rule: Doc, script: Doc, is_short: bool) -> Iterator[Doc]:
         # (1): a line-break after the rule, e.g.,
         #
         # select camel left:
@@ -396,7 +363,7 @@ class TalonFormatter:
         #
         # select camel left: user.extend_camel_left()
         #
-        if len(block.children) == 1:
+        if is_short:
             alt2 = self.format_short_command(rule, script)
         else:
             alt2 = Fail
@@ -443,12 +410,12 @@ class TalonFormatter:
             previous_line = child.end_position.line
 
     @format_lines.register
-    def _(self, node: TalonAssignment) -> Iterator[Doc]:
+    def _(self, node: TalonAssignmentStatement) -> Iterator[Doc]:
         self.assert_only_comments(node.children)
         yield self.format(node.left) // "=" // self.format(node.right) / Line
 
     @format_lines.register
-    def _(self, node: TalonExpression) -> Iterator[Doc]:
+    def _(self, node: TalonExpressionStatement) -> Iterator[Doc]:
         self.assert_only_comments(node.children)
         yield self.format(node.expression) / Line
 
@@ -465,6 +432,11 @@ class TalonFormatter:
     def _(self, node: TalonArgumentList) -> Doc:
         return ("," / Space).join(self.format_children(node.children))
 
+    @format.register
+    def _(self, node: TalonUnaryOperator) -> Doc:
+        self.assert_only_comments(node.children)
+        return self.format(node.operator) / self.format(node.right)
+
     @format.register
     def _(self, node: TalonBinaryOperator) -> Doc:
         self.assert_only_comments(node.children)
@@ -489,14 +461,9 @@ class TalonFormatter:
 
     @format.register
     def _(self, node: TalonParenthesizedExpression) -> Doc:
-        return parens(self.format(self.get_node(node.children)))
-
-    @format.register
-    def _(self, node: TalonRegexEscapeSequence) -> Doc:
-        if node.children:
-            return braces(self.format_children(node.children))
-        else:
-            return braces(Empty)
+        return parens(
+            self.format(self.get_node(node.children, node_type_name=node.type_name))
+        )
 
     @format.register
     def _(self, node: TalonSleepAction) -> Doc:
@@ -520,10 +487,6 @@ class TalonFormatter:
     def _(self, node: TalonInteger) -> Doc:
         return Text.words(node.text.strip(), collapse_whitespace=True)
 
-    @format.register
-    def _(self, node: TalonNumber) -> Doc:
-        return self.format(self.get_node(node.children))
-
     ###########################################################################
     # Format: Strings
     ###########################################################################
@@ -534,7 +497,7 @@ class TalonFormatter:
 
     @format.register
     def _(self, node: TalonInterpolation) -> Doc:
-        return self.format(self.get_node(node.children))
+        return self.format(self.get_node(node.children, node_type_name=node.type_name))
 
     @format.register
     def _(self, node: TalonString) -> Doc:
@@ -574,22 +537,25 @@ class TalonFormatter:
 
     @format.register
     def _(self, node: TalonOptional) -> Doc:
-        child = self.get_node(node.children)
+        child = self.get_node(node.children, node_type_name=node.type_name)
         return brackets(self.format(child))
 
     @format.register
     def _(self, node: TalonParenthesizedRule) -> Doc:
-        child = self.get_node(node.children)
+        child = self.get_node(node.children, node_type_name=node.type_name)
         return parens(self.format(child))
 
     @format.register
     def _(self, node: TalonRepeat) -> Doc:
-        child = self.get_node(node.children)
+        child = self.get_node(node.children, node_type_name=node.type_name)
         return self.format(child) / "*"
 
     @format.register
     def _(self, node: TalonRepeat1) -> Doc:
-        return self.format(self.get_node(node.children)) / "+"
+        return (
+            self.format(self.get_node(node.children, node_type_name=node.type_name))
+            / "+"
+        )
 
     @format.register
     def _(self, node: TalonRule) -> Doc:
@@ -626,7 +592,7 @@ class TalonFormatter:
             return Text.words(node.text, collapse_whitespace=False) / Line
 
     # Used to buffer comments encountered inline, e.g., inside a binary operator
-    _comment_buffer: list[TalonComment] = dataclasses.field(
+    _match_context_comment_buffer: list[TalonComment] = dataclasses.field(
         default_factory=list, init=False
     )
 
@@ -641,7 +607,7 @@ class TalonFormatter:
         """
         for child in children:
             if isinstance(child, TalonComment):
-                self._comment_buffer.append(child)
+                self._match_context_comment_buffer.append(child)
             elif isinstance(child, node_type):
                 yield child
             else:
@@ -652,9 +618,9 @@ class TalonFormatter:
         Get the buffered comments. Clear the buffer.
         """
         try:
-            yield from self._comment_buffer
+            yield from self._match_context_comment_buffer
         finally:
-            self._comment_buffer.clear()
+            self._match_context_comment_buffer.clear()
 
     def with_comments(self, *doclike: DocLike) -> Iterator[Doc]:
         """
@@ -672,17 +638,20 @@ class TalonFormatter:
             len(rest) == 0
         ), f"There should be no non-comment nodes, found {tuple(node.type_name for node in rest)}:\n{rest}"
 
-    def get_node(self, children: Iterable[Node]) -> Node:
+    def get_node(self, children: Iterable[Node], *, node_type_name: str) -> Node:
         """
         Get the single node that is not a comment. Store all the comments.
         """
-        return self.get_node_with_type(children, node_type=Node)
+        return self.get_node_with_type(
+            children, node_type=Node, node_type_name=node_type_name
+        )
 
     def get_node_with_type(
         self,
         children: Iterable[Union[NodeVar, TalonComment]],
         *,
         node_type: type[NodeVar],
+        node_type_name: str,
     ) -> NodeVar:
         """
         Get the single node that is not a comment, but has type NodeVar. Store all the comments.
@@ -690,5 +659,5 @@ class TalonFormatter:
         rest = tuple(self.store_comments_with_type(children, node_type=node_type))
         assert (
             len(rest) == 1
-        ), f"There should be only one non-comment child, found {tuple(node.type_name for node in rest)}:\n{rest}"
+        ), f"There should be only one non-comment child in '{node_type_name}', found {tuple(node.type_name for node in rest)}:\n{rest}"
         return next(iter(rest))
